==> ec2init-0.11-py3-none-any.whl: ec2init/plans/preinit/database_init.py <==
import os
import logging
import pathlib
import subprocess
import configparser
import shutil

from typing import List

from ec2init import Procedure
from ec2init.plans.preinit.pre_inits import BrainsPreInit
from ec2init.plans.init.start_services import StartServices
from ec2init.plans.multinode.seds import Seds
from ec2init.utils.system import SystemUtils
from ec2init.utils.cloud import get_cloud_provider
from ec2init.utils.shell import ShellUtils
from ec2init.utils.backup_information import BackupInformation
from ec2init.utils.database import databaseUtils

logger = logging.getLogger(__name__)

class MasterInitializeDatabase(BrainsPreInit):
    def __init__(self, name: str, config: configparser.ConfigParser, check_file: pathlib.Path,
        bucket: str, region: str, ldap: str, oauth: str, password_file: pathlib.Path, user_email: str,
        gateway: str, user_id: str, instance_id: str, api_key: str, cloud_provider: str, is_upgrade: bool, search_domain: str, node_id: str, user_mgm_type: str, source_version: str):
        super().__init__(name, config, check_file, bucket, region, ldap, oauth, password_file, user_email,
            gateway, user_id, instance_id, api_key, cloud_provider, is_upgrade, search_domain, node_id, user_mgm_type)
        self.source_version = source_version

    def execute(self) -> List[Procedure]:
        if BackupInformation.is_backup(self.backup_info):
            self.stop_database()
        self.fix_configuration()
        self.start_database()

        self.touch_check_file()
        return []

    def fix_configuration(self) -> None:
        # We need it only for its constants
        s = Seds(self.name, self.config, self.cloud_provider, self.search_domain, self.node_id)
        local_ip = self.get_local_ip()
        if BackupInformation.is_backup(self.backup_info):
            if self.backup_info.ip:
                s.MASTER_IP = self.backup_info.ip

        replacement = f"s/{s.MASTER_IP}/{local_ip}/g"
        logger.info("replacing in mysql config and scripts with command " + replacement)
        self.sysexec(self.sed_cmd(replacement, "/srv/hops/mysql-cluster/config.ini"))
        self.sysexec(self.sed_cmd(replacement, "/srv/hops/mysql-cluster/my.cnf"))
        total_memory = self._get_ndbd_totalmemory()
        total_memory_replacement = f"s/TotalMemoryConfig[[:space:]]*=.*$/TotalMemoryConfig={total_memory}M/g"
        self.sysexec(self.sed_cmd(total_memory_replacement, "/srv/hops/mysql-cluster/config.ini"))

        num_cpus = self._get_ndbd_cpu()
        num_cpu_replacement = f"s/NumCPUs[[:space:]]*=.*$/NumCPUs={num_cpus}/g"
        self.sysexec(self.sed_cmd(num_cpu_replacement, "/srv/hops/mysql-cluster/config.ini"))

        rondb_mgmd_connection = f"{local_ip}:1186"
        replaceMgmdConnection = f"s/RONDB_MGMD_CONNECTION/{rondb_mgmd_connection}/g"
        self.sysexec(self.sed_cmd(replaceMgmdConnection, "/srv/hops/epipe/conf/config.ini"))
        self.sysexec(self.sed_cmd(replaceMgmdConnection, "/srv/hops/epipe/conf/config-reindex.ini"))

        cmd = s.SEEK_AND_REPLACE.format("/srv/hops/mysql-cluster/ndb/scripts", replacement)
        # Unix find is a pain for subprocess
        subprocess.run(cmd, timeout=20, shell=True)

    def _get_ndbd_totalmemory(self) -> int:
        multiplier = self.config.getfloat('initplan', 'rondb_total_memory_config_multiplier', fallback=0.1)
        available_memory = SystemUtils.get_system_memory()
        datanode_memory = int(multiplier * available_memory)
        if datanode_memory > available_memory:
            datanode_memory = available_memory
        
        datanode_memory = max(4096, datanode_memory)
        return datanode_memory

    def _get_ndbd_cpu(self) -> int:
        multiplier = self.config.getfloat('initplan', 'rondb_num_cpus_multiplier', fallback=0.4)
        available_cpus = SystemUtils.get_num_cpus()
        rondb_cpus = int(multiplier * available_cpus)
        if rondb_cpus > available_cpus:
            rondb_cpus = available_cpus
        
        rondb_cpus = max(4, rondb_cpus)
        return rondb_cpus

    def start_database(self) -> None:
        self.start_ndb()
        if self.is_upgrade:
            self.restore_backups_for_upgrade()
        else:
            self.start_mysqld()
    
    def stop_database(self) -> None:
        self.stop_mysqld()
        self.stop_ndb()

    def start_ndb(self) -> None:
        self.sysexec(["systemctl", "start", "ndb_mgmd"])
        self.sysexec(["systemctl", "start", "ndbmtd"], timeout=900)

    def stop_ndb(self) -> None:
        self.sysexec(["systemctl", "stop", "ndbmtd"], timeout=900)
        self.sysexec(["systemctl", "stop", "ndb_mgmd"])

    def start_mysqld(self) -> None:
        logger.info("Starting MySQLd")
        self.sysexec(["systemctl", "start", "mysqld"], timeout=900)
        logger.info("Started MySQLd")

    def stop_mysqld(self) -> None:
        self.sysexec(["systemctl", "stop", "mysqld"], timeout=900)

    def restore_backups_for_upgrade(self) -> None:
        self.initialize_mysqld()
        self.start_mysqld()
        backups_base_dir = "/root/upgrade_db_backup"
        mysql_dumps = f"{backups_base_dir}/mysqldump"
        databases = [d.split('.sql')[0] for d in os.listdir(mysql_dumps)]
        self.restore_databases(databases, mysql_dumps)
        self.restore_users(backups_base_dir)
        local_ip = get_cloud_provider(self.cloud_provider).get_private_ip(self.config)
        self.restore_native_backup(f"{backups_base_dir}/ndb/BACKUP-1", local_ip)
        shutil.rmtree(backups_base_dir)

    def initialize_mysqld(self) -> None:
        logger.info("Initialiazing MySQLd")
        ShellUtils.rmtree("/srv/hops/mysql-cluster/mysql", keep_root=True)
        cmd = "/srv/hops/mysql/bin/mysqld --defaults-file=/srv/hops/mysql-cluster/my.cnf --initialize-insecure --explicit_defaults_for_timestamp"
        SystemUtils.execute_command(cmd.split(" "), timeout=1200)
        logger.info("Finished initializing MySQLd")
    
    def restore_databases(self, databases: List[str], base_dir: str) -> None:
        for db in databases:
            logger.info("Restoring db -> %s", db)
            full_path = f"{base_dir}/{db}.sql"
            create_cmd = ["/srv/hops/mysql-cluster/ndb/scripts/mysql-client.sh", "-e", f"create database {db}"]
            logger.info("Creating database")
            SystemUtils.execute_command(create_cmd, timeout=300)
            logger.info("Restoring SQL schema")
            restore_cmd = f"/srv/hops/mysql-cluster/ndb/scripts/mysql-client.sh {db} < {str(full_path)}"
            SystemUtils.execute_command([restore_cmd], shell=True, timeout=1200)
            logger.info("Finished restoring db -> %s", db)

    def restore_users(self, base_dir: str) -> None:
        logger.info("Restoring users")
        users_sql = f"{base_dir}/users.sql"
        restore_cmd = f"/srv/hops/mysql-cluster/ndb/scripts/mysql-client.sh < {str(users_sql)}"
        SystemUtils.execute_command([restore_cmd], shell=True, timeout=1200)
        logger.info("Finished restoring users")

    def restore_native_backup(self, backup_path: str, mgmd_ip: str) -> None:
        logger.info("Restoring NDB backup")
        ndb_restore_cmd = f"/srv/hops/mysql/bin/ndb_restore --ndb-connectstring={mgmd_ip}:1186 --nodeid=1 --backupid=1 --restore-data --backup-path={backup_path}"
        SystemUtils.execute_command(ndb_restore_cmd.split(" "), timeout=1200)
        logger.info("Finished restoring NDB backup")

class LocalhostInitializeDatabase(MasterInitializeDatabase):
    def __init__(self, name: str, config: configparser.ConfigParser, check_file: pathlib.Path,
        bucket: str, region: str, ldap: str, oauth: str, password_file: pathlib.Path, user_email: str,
        gateway: str, user_id: str, instance_id: str, api_key: str, cloud_provider: str, is_upgrade: bool, search_domain: str,
         node_id: str, user_mgm_type: str, source_version: str):
        super().__init__(name, config, check_file, bucket, region, ldap, oauth, password_file, user_email,
            gateway, user_id, instance_id, api_key, cloud_provider, is_upgrade, search_domain, node_id, user_mgm_type, source_version)

    def fix_configuration(self) -> None:
        pass


class MasterWithNdbInitializeDatabase(MasterInitializeDatabase):
    def __init__(self, name: str, config: configparser.ConfigParser, check_file: pathlib.Path,
        bucket: str, region: str, ldap: str, oauth: str, password_file: pathlib.Path, user_email: str,
        gateway: str, user_id: str, instance_id: str, api_key: str, cloud_provider: str,
        mgmd_ip: str, is_upgrade: bool, search_domain: str, node_id: str, user_mgm_type: str, source_version: str):
        super().__init__(name, config, check_file, bucket, region, ldap, oauth, password_file, user_email,
            gateway, user_id, instance_id, api_key, cloud_provider, is_upgrade, search_domain, node_id, user_mgm_type, source_version)
        self.mgmd_ip = mgmd_ip

        
    def fix_configuration(self) -> None:
        # We need it only for its constants
        s = Seds(self.name, self.config, self.cloud_provider, self.search_domain, self.node_id)
        ## TODO We need to deal with backup
        if BackupInformation.is_backup(self.backup_info):
            if self.backup_info.ip:
                s.MASTER_IP = self.backup_info.ip

        replacement = f"s/{s.MASTER_IP}/{self.mgmd_ip}/g"
        logger.info("replacing in mysql config and scripts with command " + replacement)
        self.sysexec(self.sed_cmd(replacement, "/srv/hops/mysql-cluster/config.ini"))
        self.sysexec(self.sed_cmd(replacement, "/srv/hops/mysql-cluster/my.cnf"))

        self.sysexec(self.sed_cmd(replacement, "/srv/hops/hadoop/etc/hadoop/ndb.props"))

        ## TODO: This must be multiple addresses
        rondb_mgmd_connection = f"{self.mgmd_ip}:1186"
        replaceMgmdConnection = f"s/RONDB_MGMD_CONNECTION/{rondb_mgmd_connection}/g"
        self.sysexec(self.sed_cmd(replaceMgmdConnection, "/srv/hops/epipe/conf/config.ini"))
        self.sysexec(self.sed_cmd(replaceMgmdConnection, "/srv/hops/epipe/conf/config-reindex.ini"))

        cmd = s.SEEK_AND_REPLACE.format("/srv/hops/mysql-cluster/ndb/scripts", replacement)
        # Unix find is a pain for subprocess
        subprocess.run(cmd, timeout=20, shell=True)

        # Remove NDB from kagent services
        kagent_conf_dir = self.config.get(
            'general', 'kagent_conf_dir', fallback="/srv/hops/kagent/etc")
        services_file = pathlib.Path(kagent_conf_dir, "services")
        services = configparser.ConfigParser()
        services.read(str(services_file))

        if 'NDB-ndb_mgmd' in services:
            del services['NDB-ndb_mgmd']
        if 'NDB-ndbmtd' in services:
            del services['NDB-ndbmtd']
        
        with open(str(services_file), 'w') as fd:
            services.write(fd)

    def start_database(self) -> None:
        # It's fine to pass empty arguments, we only need the
        # helper methods to manage systemd units
        ss = StartServices(MasterWithNdbInitializeDatabase.__name__,
            "", "", self.node_id, "", "", self.config, ("", ""), "", self.cloud_provider, self.search_domain)

        logger.info("Disabling ndb_mgmd")
        ss._stop_service("ndb_mgmd")
        ss._disable_service("ndb_mgmd")

        logger.info("Disabling ndbmtd")
        ss._stop_service("ndbmtd")
        ss._disable_service("ndbmtd")

        ss._stop_service("mysqld")
        # Initialize MySQLd
        self.initialize_mysqld()

        # Start again MySQLd
        self.start_mysqld()
            
        if self.is_upgrade:
            # In a RonDB cluster, the backup is taken from the MySQL node of the RonDB cluster - not the Head node
            # The view below is created *only* in the Head node thus, does not exist in the backup and we need to re-create it
            databaseUtils.run_command(["hopsworks", "-e", "DROP VIEW IF EXISTS `users_groups`"])
            databaseUtils.run_command(["hopsworks", "-e", "CREATE ALGORITHM=UNDEFINED VIEW `users_groups` AS select `u`.`username` AS `username`,`u`.`password` AS `password`,`u`.`secret` AS `secret`,`u`.`email` AS `email`,`g`.`group_name` AS `group_name` from ((`user_group` `ug` join `users` `u` on((`u`.`uid` = `ug`.`uid`))) join `bbc_group` `g` on((`g`.`gid` = `ug`.`gid`)))"])
            
            # Flyway table is using innoDB by default for older clusters (2.2 and 2.4) running RonDB
            # alter this table and insert the corresponding initial table row to the flyway schema table
            if len(SystemUtils.execute_mysql_statement("SHOW TABLES LIKE 'flyway_schema_history'", "hopsworks")) == 0:
                SystemUtils.execute_command("/srv/hops/domains/domain1/flyway/flyway baseline".split(" "), timeout=600)
                databaseUtils.run_command(["hopsworks", "-e", "truncate flyway_schema_history"])
                databaseUtils.run_command(["hopsworks", "-e", "ALTER TABLE flyway_schema_history engine = 'ndb'"])
                if self.source_version == "2.2.0":
                    databaseUtils.run_command(["hopsworks", "-e", "INSERT INTO hopsworks.flyway_schema_history VALUES(1, '2.2.0', 'initial tables', 'SQL', 'V2.2.0__initial_tables.sql', 334646086, 'kthfs', '2021-12-13 12:35:06', 5, 1)"])
                elif self.source_version == "2.4.0":
                    databaseUtils.run_command(["hopsworks", "-e", "INSERT INTO hopsworks.flyway_schema_history VALUES(1, '2.4.0', 'initial tables', 'SQL', 'V2.4.0__initial_tables.sql', -1798611666, 'kthfs', '2021-12-13 12:35:06', 5, 1)"])
        else:
            # Restore the backup
            databases = ["airflow", "glassfish_timers", "hops", "hopsworks",
                     "metastore"]

            logger.info("Started restoring databases")
            backup_base = "/root/db_backup"
            mysqldumps = f"{backup_base}/mysqldump"
            ndb_backup = f"{backup_base}/ndb/BACKUP-1"

            self.restore_databases(databases, mysqldumps)
            self.restore_users(mysqldumps)
            self.restore_native_backup(ndb_backup, self.mgmd_ip)

==> ec2init-0.17-py3-none-any.whl: ec2init/plans/preinit/database_init.py <==
import os
import logging
import pathlib
import subprocess
import configparser
import shutil

from typing import List

from ec2init import Procedure
from ec2init.plans.preinit.pre_inits import BrainsPreInit
from ec2init.plans.init.start_services import StartServices
from ec2init.plans.multinode.seds import Seds
from ec2init.utils.bucket_configuration import BucketConfiguration
from ec2init.utils.system import SystemUtils
from ec2init.utils.cloud import get_cloud_provider
from ec2init.utils.shell import ShellUtils
from ec2init.utils.offline_backup_information import OffLineBackupInformation
from ec2init.utils.database import databaseUtils
from ec2init.utils.constants import Scripts

logger = logging.getLogger(__name__)

class MasterInitializeDatabase(BrainsPreInit):

    def __init__(self, name: str, config: configparser.ConfigParser, check_file: pathlib.Path,
        bucket_configuration: BucketConfiguration, region: str, ldap: str, oauth: str, password_file: pathlib.Path, user_email: str,
        gateway: str, user_id: str, instance_id: str, api_key: str, cloud_provider: str, is_upgrade: bool, search_domain: str, node_id: str, user_mgm_type: str, 
        create_pwd_user: bool, source_version: str, is_backup: bool, offline_backup_info: OffLineBackupInformation, cluster_domain_prefix: str):
        super().__init__(name, config, check_file, bucket_configuration, region, ldap, oauth, password_file, user_email,
            gateway, user_id, instance_id, api_key, cloud_provider, is_upgrade, search_domain, node_id, user_mgm_type, create_pwd_user, is_backup, offline_backup_info, cluster_domain_prefix)
        self.source_version = source_version

    def execute(self) -> List[Procedure]:
        if self.is_backup:
            self.stop_database()
        self.fix_configuration()
        self.start_database()

        self.touch_check_file()
        return []

    def fix_configuration(self) -> None:
        # We need it only for its constants
        s = Seds(self.name, self.config, self.cloud_provider, self.search_domain, self.node_id, self.gateway, self.user_id, self.instance_id, self.api_key)
        local_ip = self.get_local_ip()
        if OffLineBackupInformation.is_backup(self.offline_backup_info):
            if self.offline_backup_info.ip:
                s.MASTER_IP = self.offline_backup_info.ip

        replacement = f"s/{s.MASTER_IP}/{local_ip}/g"
        logger.info("replacing in mysql config and scripts with command " + replacement)
        self.sysexec(self.sed_cmd(replacement, "/srv/hops/mysql-cluster/config.ini"))
        self.sysexec(self.sed_cmd(replacement, "/srv/hops/mysql-cluster/my.cnf"))
        total_memory = self._get_ndbd_totalmemory()
        total_memory_replacement = f"s/TotalMemoryConfig[[:space:]]*=.*$/TotalMemoryConfig={total_memory}M/g"
        self.sysexec(self.sed_cmd(total_memory_replacement, "/srv/hops/mysql-cluster/config.ini"))

        num_cpus = self._get_ndbd_cpu()
        num_cpu_replacement = f"s/NumCPUs[[:space:]]*=.*$/NumCPUs={num_cpus}/g"
        self.sysexec(self.sed_cmd(num_cpu_replacement, "/srv/hops/mysql-cluster/config.ini"))

        rondb_mgmd_connection = f"{local_ip}:1186"
        replaceMgmdConnection = f"s/RONDB_MGMD_CONNECTION/{rondb_mgmd_connection}/g"
        self.sysexec(self.sed_cmd(replaceMgmdConnection, "/srv/hops/epipe/conf/config.ini"))
        self.sysexec(self.sed_cmd(replaceMgmdConnection, "/srv/hops/epipe/conf/config-reindex.ini"))

        cmd = s.SEEK_AND_REPLACE.format("/srv/hops/mysql-cluster/ndb/scripts", replacement)
        # Unix find is a pain for subprocess
        subprocess.run(cmd, timeout=20, shell=True)

    def _get_ndbd_totalmemory(self) -> int:
        multiplier = self.config.getfloat('initplan', 'rondb_total_memory_config_multiplier', fallback=0.1)
        available_memory = SystemUtils.get_system_memory()
        datanode_memory = int(multiplier * available_memory)
        if datanode_memory > available_memory:
            datanode_memory = available_memory
        
        datanode_memory = max(4096, datanode_memory)
        return datanode_memory

    def _get_ndbd_cpu(self) -> int:
        multiplier = self.config.getfloat('initplan', 'rondb_num_cpus_multiplier', fallback=0.4)
        available_cpus = SystemUtils.get_num_cpus()
        rondb_cpus = int(multiplier * available_cpus)
        if rondb_cpus > available_cpus:
            rondb_cpus = available_cpus
        
        rondb_cpus = max(4, rondb_cpus)
        return rondb_cpus

    def start_database(self) -> None:
        self.start_ndb()
        self.start_mysqld()
    
    def stop_database(self) -> None:
        self.stop_mysqld()
        self.stop_ndb()

    def start_ndb(self) -> None:
        self.sysexec(["systemctl", "start", "ndb_mgmd"])
        self.sysexec(["systemctl", "start", "ndbmtd"], timeout=900)

    def stop_ndb(self) -> None:
        self.sysexec(["systemctl", "stop", "ndbmtd"], timeout=900)
        self.sysexec(["systemctl", "stop", "ndb_mgmd"])

    def start_mysqld(self) -> None:
        logger.info("Starting MySQLd")
        post_exception_action = lambda: SystemUtils.execute_command(["systemctl", "stop", "mysqld"], timeout=500)
        SystemUtils.execute_command_with_retry(["systemctl", "start", "mysqld"], post_exception_action=post_exception_action, timeout=900)
        logger.info("Started MySQLd")

    def stop_mysqld(self) -> None:
        self.sysexec(["systemctl", "stop", "mysqld"], timeout=900)

    def initialize_mysqld(self) -> None:
        logger.info("Initialiazing MySQLd")
        ShellUtils.rmtree("/srv/hops/mysql-cluster/mysql", keep_root=True)
        cmd = "/srv/hops/mysql/bin/mysqld --defaults-file=/srv/hops/mysql-cluster/my.cnf --initialize-insecure --explicit_defaults_for_timestamp"
        SystemUtils.execute_command(cmd.split(" "), timeout=1200)
        logger.info("Finished initializing MySQLd")
    
    def restore_databases(self, databases: List[str], base_dir: str) -> None:
        for db in databases:
            logger.info("Restoring db -> %s", db)
            full_path = f"{base_dir}/{db}.sql"
            create_cmd = [Scripts.MYSQL_CLIENT, "-e", f"create database {db}"]
            logger.info("Creating database")
            SystemUtils.execute_command(create_cmd, timeout=300)
            logger.info("Restoring SQL schema")
            restore_cmd = f"{Scripts.MYSQL_CLIENT} {db} < {str(full_path)}"
            SystemUtils.execute_command([restore_cmd], shell=True, timeout=1200)
            logger.info("Finished restoring db -> %s", db)

    def restore_users(self, base_dir: str) -> None:
        logger.info("Restoring users")
        users_sql = f"{base_dir}/users.sql"
        restore_cmd = f"{Scripts.MYSQL_CLIENT} < {str(users_sql)}"
        SystemUtils.execute_command([restore_cmd], shell=True, timeout=1200)
        logger.info("Finished restoring users")

    def restore_native_backup(self, backup_path: str, mgmd_ip: str) -> None:
        logger.info("Restoring NDB backup")
        ndb_restore_cmd = f"/srv/hops/mysql/bin/ndb_restore --ndb-connectstring={mgmd_ip}:1186 --nodeid=1 --backupid=1 --restore-data --backup-path={backup_path}"
        SystemUtils.execute_command(ndb_restore_cmd.split(" "), timeout=1200)
        logger.info("Finished restoring NDB backup")

class MasterWithNdbInitializeDatabase(MasterInitializeDatabase):
    def __init__(self, name: str, config: configparser.ConfigParser, check_file: pathlib.Path,
        bucket_configuration: BucketConfiguration, region: str, ldap: str, oauth: str, password_file: pathlib.Path, user_email: str,
        gateway: str, user_id: str, instance_id: str, api_key: str, cloud_provider: str,
        mgmd_ip: str, is_upgrade: bool, search_domain: str, node_id: str, user_mgm_type: str, create_pwd_user: bool, source_version: str, is_backup: bool, offline_backup_info: OffLineBackupInformation,
        cluster_domain_prefix: str):
        super().__init__(name, config, check_file, bucket_configuration, region, ldap, oauth, password_file, user_email,
            gateway, user_id, instance_id, api_key, cloud_provider, is_upgrade, search_domain, node_id, user_mgm_type, create_pwd_user, source_version, is_backup, offline_backup_info,
            cluster_domain_prefix)
        self.mgmd_ip = mgmd_ip

        
    def fix_configuration(self) -> None:
        # We need it only for its constants
        s = Seds(self.name, self.config, self.cloud_provider, self.search_domain, self.node_id, self.gateway, self.user_id, self.instance_id, self.api_key)
        ## TODO We need to deal with backup
        if OffLineBackupInformation.is_backup(self.offline_backup_info):
            if self.offline_backup_info.ip:
                s.MASTER_IP = self.offline_backup_info.ip

        replacement = f"s/{s.MASTER_IP}/{self.mgmd_ip}/g"
        logger.info("replacing in mysql config and scripts with command " + replacement)
        self.sysexec(self.sed_cmd(replacement, "/srv/hops/mysql-cluster/config.ini"))
        self.sysexec(self.sed_cmd(replacement, "/srv/hops/mysql-cluster/my.cnf"))

        self.sysexec(self.sed_cmd(replacement, "/srv/hops/hadoop/etc/hadoop/ndb.props"))

        ## TODO: This must be multiple addresses
        rondb_mgmd_connection = f"{self.mgmd_ip}:1186"
        replaceMgmdConnection = f"s/RONDB_MGMD_CONNECTION/{rondb_mgmd_connection}/g"
        self.sysexec(self.sed_cmd(replaceMgmdConnection, "/srv/hops/epipe/conf/config.ini"))
        self.sysexec(self.sed_cmd(replaceMgmdConnection, "/srv/hops/epipe/conf/config-reindex.ini"))

        cmd = s.SEEK_AND_REPLACE.format("/srv/hops/mysql-cluster/ndb/scripts", replacement)
        # Unix find is a pain for subprocess
        subprocess.run(cmd, timeout=20, shell=True)

        # Remove NDB from kagent services
        kagent_conf_dir = self.config.get(
            'general', 'kagent_conf_dir', fallback="/srv/hops/kagent/etc")
        services_file = pathlib.Path(kagent_conf_dir, "services")
        services = configparser.ConfigParser()
        services.read(str(services_file))

        if 'NDB-ndb_mgmd' in services:
            del services['NDB-ndb_mgmd']
        if 'NDB-ndbmtd' in services:
            del services['NDB-ndbmtd']
        
        with open(str(services_file), 'w') as fd:
            services.write(fd)

    def start_database(self) -> None:
        logger.info("Disabling ndb_mgmd")
        StartServices.stop_service("ndb_mgmd")
        StartServices.disable_service("ndb_mgmd")

        logger.info("Disabling ndbmtd")
        StartServices.stop_service("ndbmtd")
        StartServices.disable_service("ndbmtd")

        StartServices.stop_service("mysqld")
        # Initialize MySQLd
        self.initialize_mysqld()

        # Start again MySQLd
        self.start_mysqld()
            
        if self.is_backup or self.is_upgrade:
            # In a RonDB cluster, the backup is taken from the MySQL node of the RonDB cluster - not the Head node
            # The view below is created *only* in the Head node thus, does not exist in the backup and we need to re-create it
            databaseUtils.run_command(["hopsworks", "-e", "DROP VIEW IF EXISTS `users_groups`"])
            databaseUtils.run_command(["hopsworks", "-e", "CREATE ALGORITHM=UNDEFINED VIEW `users_groups` AS select `u`.`username` AS `username`,`u`.`password` AS `password`,`u`.`secret` AS `secret`,`u`.`email` AS `email`,`g`.`group_name` AS `group_name` from ((`user_group` `ug` join `users` `u` on((`u`.`uid` = `ug`.`uid`))) join `bbc_group` `g` on((`g`.`gid` = `ug`.`gid`)))"])

        if self.is_upgrade:
            # Flyway table is using innoDB by default for older clusters (2.2 and 2.4) running RonDB
            # alter this table and insert the corresponding initial table row to the flyway schema table
            if len(SystemUtils.execute_mysql_statement("SHOW TABLES LIKE 'flyway_schema_history'", "hopsworks")) == 0:
                SystemUtils.execute_command("/srv/hops/domains/domain1/flyway/flyway baseline".split(" "), timeout=600)
                databaseUtils.run_command(["hopsworks", "-e", "truncate flyway_schema_history"])
                databaseUtils.run_command(["hopsworks", "-e", "ALTER TABLE flyway_schema_history engine = 'ndb'"])
                if self.source_version == "2.2.0":
                    databaseUtils.run_command(["hopsworks", "-e", "INSERT INTO hopsworks.flyway_schema_history VALUES(1, '2.2.0', 'initial tables', 'SQL', 'V2.2.0__initial_tables.sql', 334646086, 'kthfs', '2021-12-13 12:35:06', 5, 1)"])
                elif self.source_version == "2.4.0":
                    databaseUtils.run_command(["hopsworks", "-e", "INSERT INTO hopsworks.flyway_schema_history VALUES(1, '2.4.0', 'initial tables', 'SQL', 'V2.4.0__initial_tables.sql', -1798611666, 'kthfs', '2021-12-13 12:35:06', 5, 1)"])

        if not self.is_upgrade and not self.is_backup:
            ## Initial cluster start
            
            logger.info("Started restoring databases")
            backup_base = "/root/db_backup"
            backup_id = 1
            backup_path = f"{backup_base}/initial-backup-{backup_id}"

            SystemUtils.execute_command(f"tar xzf {backup_base}/initial-backup-{backup_id}.tar.gz -C {backup_base}".split(" "))
            self.restore_database_native(backup_path, self.mgmd_ip)
            self.restore_database_mysql_meta(backup_path)

    def restore_database_native(self, backup_path: str, mgmd_ip: str) -> None:
        
        logger.info("Restoring RonDB native database backup")
        logger.info("Restore meta")
        cmd = f"{Scripts.DATABASE_RESTORE_BACKUP} ndb-restore -p {backup_path} -n 1 -b 1 -c {mgmd_ip}:1186 -d -m META"
        SystemUtils.execute_command(cmd.split(" "), timeout=1200)

        logger.info("Restore data")
        cmd = f"{Scripts.DATABASE_RESTORE_BACKUP} ndb-restore -p {backup_path} -n 1 -b 1 -c {mgmd_ip}:1186 -m DATA"
        SystemUtils.execute_command(cmd.split(" "), timeout=1200)

        logger.info("Rebuild indexes")
        cmd = f"{Scripts.DATABASE_RESTORE_BACKUP} ndb-restore -p {backup_path} -n 1 -b 1 -c {mgmd_ip}:1186 -m REBUILD-INDEXES"
        SystemUtils.execute_command(cmd.split(" "), timeout=1200)
        logger.info("Finished restoring RonDB native backup")

    def restore_database_mysql_meta(self, backup_path: str) -> None:
        logger.info("Restore MySQL metadata")
        restore_cmd = f"{Scripts.DATABASE_RESTORE_BACKUP} restore-schema -p {backup_path}"
        SystemUtils.execute_command(restore_cmd.split(" "), timeout=300)
        logger.info("Finished restoring MySQL metadata")
     


class SecondaryWithMySQLd(MasterWithNdbInitializeDatabase):
    def __init__(self, name: str, config: configparser.ConfigParser, check_file: pathlib.Path,
        bucket_configuration: BucketConfiguration, region: str, ldap: str, oauth: str, password_file: pathlib.Path, user_email: str,
        gateway: str, user_id: str, instance_id: str, api_key: str, cloud_provider: str,
        mgmd_ip: str, is_upgrade: bool, search_domain: str, node_id: str, user_mgm_type: str, create_pwd_user: bool, source_version: str, is_backup: bool, offline_backup_info: OffLineBackupInformation, cluster_domain_prefix: str,):

        super().__init__(name, config, check_file, bucket_configuration, region, ldap, oauth, password_file, user_email,
        gateway, user_id, instance_id, api_key, cloud_provider, mgmd_ip, is_upgrade, search_domain, node_id, user_mgm_type, create_pwd_user, source_version, is_backup, offline_backup_info, cluster_domain_prefix)
        
    def stop_database(self) -> None:
        pass

    def start_database(self) -> None:
        StartServices.stop_service("mysqld")
        # Initialize MySQLd
        self.initialize_mysqld()

        # Start again MySQLd
        self.start_mysqld()
